/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file lka.h
 *
 * The lka is the data structure that keeps track of all of the users on the
 * network. This class is implemented as a singleton so there can only be one
 * instance during the lifetime of the entire program. Some of the calls have
 * c-style type signatures because they are meant to be closely tied with the
 * plugin-api.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef _P4_LKA_H
#define _P4_LKA_H

#include "protocol.h"

#include <pthread.h>
#include <sys/types.h>
#include <semaphore.h>
#include <utility>
#include <set>
#include <map>
#include <vector>
#include <algorithm>
using namespace std;

typedef pair<lka_entry, plugin_info*> lka_pair;

/************************************************************************
 * CLASS clka
 ***********************************************************************/
class clka {

  public:

  /** instance
   * 
   * Returns the only instance of clka.  This implements the singleton
   * pattern.
   */
  static clka &instance();

  /** wipe
   *
   * Removes all entries from the lka.  Frees usernames and plugin_info
   * lists.
   */
  uint wipe();
  
  /** populate
   *
   * Populate with data from an lka_ack.
   */
  uint populate(p4_lka_ack *ack);

  /** package
   *
   * Packages up the lka to be sent as an lka_ack.  NOTE: The p4_lka_ack must
   * be deleted after use.
   *
   * Returns a chunk of memory that has the following format:
   * --------------------------------------------------------------------------
   * | p4_lka_ack | lka_entry | lka_entry |...| plugin_info | plugin_info |...|
   * --------------------------------------------------------------------------
   *
   * The p4_lka_ack tells how many lka_entries there are.  Each lka_entry
   * tells how many plugin_infos it has.  That means that if the first 
   * lka_entry has 3 plugins, the first three plugin_infos belong to that
   * that entry.  The second entry gets its plugin_infos starting at the
   * fourth and so on.
   */
  p4_lka_ack *package(ulong &package_size);

  /** add_user
   *
   * Put a new user in the lka.  If user is already present, 
   * overwrites data for that user with new values.
   */
  uint add(const char *username, ulong addr, ushort port, ushort num_plugins,
           const plugin_info *plugins);

  /** remove_user
   *
   * Removes user from lka data structure.  Frees plugin_info and the 
   * user name. Returns non-zero if error.
   */
  uint remove(const char *username);

  /** get_address
   *
   * Writes ip address and port for user into addr and port.
   * Returns non-zero if error.
   */
  uint get_address(const char *username, ulong &addr, ushort &port) const;

  /** get_num_users
   *
   * Returns the number of users running plugin.
   *
   * The overloaded version doesn't take a plugin id, and returns the total
   * number of users connected to the network.
   */
  uint get_num_users(ushort plugin_id, uint &num_users) const;
  uint get_num_users(uint &num_users) const;


  /** get_users
   *
   * Writes all of the users running plugin into the users array.  This
   * array of size num_users (each 20 byte arrays) is allocated by the 
   * caller.  The actual number of users written into the array is written
   * to num_written.
   *
   * The overloaded version doesn't take a plugin id, and returns all users
   * connected to the network.
   */
  uint get_users(ushort plugin_id, char **users, uint num_users,
                 uint &num_written) const;
  uint get_users(char **users, uint num_users, uint &num_written) const;

  /** get_num_plugins
   *
   * Gets the number of active plugins for user and writes it into num_plugins.
   */
  uint get_num_plugins(const char *username, ushort &num_plugins) const;

  /** get_plugins
   *
   * Gets the active plugins for user and writes at most num_plugins 
   * worth of them to plugin_ids.  The actual number written is placed
   * in num_written.
   */
  uint get_plugins(const char *username, ushort *plugin_ids,
                   ushort num_plugins, ushort &num_written) const;

  /** update_local_user
   *
   * Repopulates the local user's entry with the current plugins. (Doesn't
   * change the local user's username, address, or port.)
   *
   * Called from creceiveq::plugin_loaded and plugin_stopped.
   */
  void update_local_user();

  /** random_user
   *
   * Gets a random username (except for the local user) from the lka.  
   * If there are not at least min_entries in the lka, returns an error.
   * Remember that the local user will be in the lka so take that into 
   * account when calculating min_entries.
   */
  uint random_user(int min_entries, string &username) const;

  /** struct lka_less
   *
   * Comparison functor for the LKA set.
   */
  struct lka_less {
    public:
    bool operator() (lka_pair * const &p1, lka_pair * const &p2) const
    {
      return (strcmp(p1->first.username, p2->first.username) < 0);
    }
  };

 protected:
  
  /** ctor
   *
   * Create new empty lka.  We must add users before the lka will have
   * any use.
   */
  clka();

  /** dtor
   *
   * Destructor, frees all entries.
   */
  ~clka();
  
  private:

  /** lookup
   *
   * Looks up a user in the lka.
   */
  set<lka_pair *, lka_less>::iterator lookup(const char *username);
  set<lka_pair *, lka_less>::const_iterator lookup(const char *username) const;
    
  /** plugin_list_add, plugin_list_remove
   *
   * Function objects used with for_each. Used to iterator over an
   * array of plugin ids and adds the lka_pair * to the plugin_lists set
   * under that plugin id.
   */
  struct plugin_list_add {
    public:
    plugin_list_add(lka_pair *p) : lka_p(p) {}

    void operator() (plugin_info pi)
    {
      clka::instance().plugin_lists[pi.id].push_back(lka_p);
    }

    private:
      lka_pair *lka_p;
  };

  struct plugin_list_remove {
    public:
    plugin_list_remove(lka_pair *p) : lka_p(p) {}

    void operator() (plugin_info pi)
    {
      vector<lka_pair*>::iterator it = 
        find(clka::instance().plugin_lists[pi.id].begin(), 
             clka::instance().plugin_lists[pi.id].end(),
             lka_p);
      assert(it != clka::instance().plugin_lists[pi.id].end());
      clka::instance().plugin_lists[pi.id].erase(it);
    }

    private:
    lka_pair *lka_p;
  };

  friend struct plugin_list_add;
  friend struct plugin_list_remove;


  /************************************************************************
   * CONSTANTS AND MEMBERS
   ***********************************************************************/

  /**
   * Stores entries for each user with address, port, # plugins, and
   * a plugin_info for each plugin.
   */
  set<lka_pair *, lka_less> entries;

  mutable pthread_rwlock_t rwlock;

  /** 
   * There is one vector for every plugin.  The vector contains
   * pointers to the structures actually in entries.
   */
  map<ushort, vector<lka_pair *> > plugin_lists;

  // sem_t lka_sem;

  // static const int kmax_readers;
};

/***************************************************************************
 * INLINE METHODS
 **************************************************************************/

inline clka &clka::instance()
{
  static clka lka;

  return lka;
}

#endif // _P4_LKA_H
